#include <Wire.h> 
#include <Serial.h>
#include <LCDI2C4Bit.h>
#include <dht.h>
#include <OneWire.h>
#include <DallasTemperature.h>


// port mappings
#define DHT22_PIN 4
#define ONE_WIRE_PIN 6
#define PELTIER_PIN 5
#define ENCODER_SA_PIN 2
#define ENCODER_SB_PIN 3

// constants
#define dht_delay 3000
#define ds_delay 1000
#define display_delay 1000 // switch between temp/humidity and peltier% / dew point
#define display_modes 3
#define Encoder_step 0.5
#define temp_init 25.0
#define Kp -0.25
#define Ki -0.001
#define Kd -0.
#define max_duty 0.8 // maximum duty cycle for the peltier element

float pid_Kp = Kp;
float pid_Ki = Ki;
float pid_Kd = Kd;
const float pid_integral_max = 0.;
const float pid_integral_min = -1000.;

// global objects
LCDI2C4Bit lcd = LCDI2C4Bit(0x27,2,40);
OneWire oneWire(ONE_WIRE_PIN);
DallasTemperature sensors(&oneWire);
dht DHT;

// global variables
unsigned long current_time=0;
unsigned long dht_timer=0;
unsigned long ds_timer=0;
unsigned long display_timer=0;
float dewpoint;
float temp_target;
float temp_target_limited;
float temp_cmos;
float temp_ambient;
float humidity;
float peltier_duty=100.; // peltier duty cycle
boolean dht_update = true;
boolean ds_update = true;
boolean target_update = true;
byte display_mode = 0;
float pid_integral = 0;
float pid_error_last;
byte peltier_mode = 0; // 0=dewpoint,1=cooling,2=maintain

// custom chars for lcd
byte celsiusChar[8] = {
	0b11000,
	0b11000,
	0b00000,
	0b00111,
	0b00100,
	0b00100,
	0b00111,
	0b00000
};
byte playChar[8] = {
	0b00000,
	0b01000,
	0b01100,
	0b01110,
	0b01100,
	0b01000,
	0b00000,
	0b00000
};
byte pauseChar[8] = {
	0b00000,
	0b01010,
	0b01010,
	0b01010,
	0b01010,
	0b01010,
	0b00000,
	0b00000
};
byte circleChar[8] = {
	0b00000,
	0b10001,
	0b00100,
	0b01110,
	0b01110,
	0b00100,
	0b10001,
	0b00000
};

void lcd_print_float(float number, int width=5, int decimals=1)
{
  char tmp[5];
  dtostrf(number,width,decimals,tmp);
  lcd.print(tmp);
}

float dewPoint(float celsius, float humidity)
{
	float a = 17.271;
	float b = 237.7;
	float temp = (a * celsius) / (b + celsius) + log(humidity*0.01);
	float t = (b * temp) / (a - temp);
	return t+1;
}

void setup()
{
  // setup serial port
  Serial.begin(115200);

  // start I2C bus
  Wire.begin();
  lcd.init();               // initialize the lcd 
  lcd.createChar(3,celsiusChar);
  lcd.createChar(0,pauseChar);
  lcd.createChar(1,playChar);
  lcd.createChar(2,circleChar);
  lcd.clear();
  lcd.home();
  lcd.backLight(true);
  lcd.print(" 1100D cooling");
 
  // peltier
  pinMode(PELTIER_PIN, OUTPUT);
 
  // rotary encoder
  attachInterrupt(1, Encoder_int, FALLING);        //interrupts: numbers 0 (on digital pin 2) and 1 (on digital pin 3).
  pinMode(ENCODER_SA_PIN, INPUT);
//  digitalWrite(ENCODER_SA_PIN, HIGH);
  pinMode(ENCODER_SB_PIN, INPUT);
//  digitalWrite(ENCODER_SB_PIN, HIGH);

  // one-wire temperature sensor
  sensors.begin();
  //sensors.setResolution(12); // stored to eeprom, only needed once
  ReadSensors(); // do initial reading of sensors
  delay(3000);
  ReadSensors();
  sensors.setWaitForConversion(false); // use library in async mode

  // set target_temp to dew-point + 1 degree C
  //temp_target = (float)((int)dewpoint+0.5);
  temp_target = temp_init;
  // # clear lcd
  lcd.clear();
  
}

void loop()
{
  current_time = millis();
  ReadSensors();
  PIcontroller();
  Display();
  SendSerial();
  ReadSerial();
  delay(1);
}

void ReadSensors()
{
  if (current_time - dht_timer >= dht_delay)
  {
    int chk = DHT.read22(DHT22_PIN);
    temp_ambient = DHT.temperature;
    humidity = DHT.humidity;
    dewpoint = dewPoint(DHT.temperature,DHT.humidity);
    dht_timer = current_time;
    dht_update = true;
  }
  else
  {
    dht_update = false;    
  }
    
  if (current_time - ds_timer >= ds_delay-750) // ds18b20 uses 750ms for temp measurement in 12bit mode
  {
    sensors.requestTemperatures(); // Send the command to get temperatures    
  }
  if (current_time - ds_timer >= ds_delay)
  {
    temp_cmos = sensors.getTempCByIndex(0);
    ds_timer = current_time;
    ds_update = true;
  }
  else
  {
    ds_update = false;
  }
}

void Display()
{
  if (current_time - display_timer >= display_delay)
  {
    display_mode += 1;
    if (display_mode > display_modes)
      display_mode = 0;
    display_timer = current_time;
  }
  
  if (dht_update && (display_mode == 0))
  {
    lcd.setCursor(0,0);
    lcd.print("A");
    lcd_print_float(temp_ambient,3,0);
    lcd.write((uint8_t)3);
  }

  if (dht_update && (display_mode == 1))
  {
    lcd.setCursor(0,0);
    lcd.print("D");
    lcd_print_float(dewpoint,3,0);
    lcd.write((uint8_t)3);
  }

  if (dht_update && (display_mode == 2))
  {
    lcd.setCursor(0,0);
    lcd.print("H");
    lcd_print_float(humidity,3,0);
    lcd.print("%");
  }

  if (ds_update)
  {
    lcd.setCursor(0,1);
    lcd.print("P");
    lcd_print_float(peltier_duty,3,0);
    lcd.print("%");
    lcd.setCursor(9,1);
    lcd.print("S");
    lcd_print_float(temp_cmos);
    lcd.write((uint8_t)3);
    
    lcd.setCursor(7,0);
    lcd.write((uint8_t)peltier_mode);
  }
 
  if (target_update)
  {
    lcd.setCursor(9,0);
    lcd.print("T");
    lcd_print_float(temp_target);
    lcd.write((uint8_t)3);
    target_update = false;
  }
}

void SendHeader()
{
    Serial.print('time(ms)\t');
    Serial.print('target(C)\t');
    Serial.print('target_limited(C)\t');
    Serial.print('cmos\t');
    Serial.print('ambient\t');
    Serial.print('humidity\n');
}

void SendSerial()
{
  if (ds_update)
  {
    Serial.print(current_time, DEC);
    Serial.print('\t');
    Serial.print(temp_target, DEC);
    Serial.print('\t');
    Serial.print(temp_target_limited, DEC);
    Serial.print('\t');
    Serial.print(temp_cmos, DEC);
    Serial.print('\t');
    Serial.print(temp_ambient, DEC);
    Serial.print('\t');
    Serial.print(humidity, DEC);
    Serial.print('\t');
    Serial.print(peltier_duty, DEC);
    Serial.print('\t');
    Serial.print(pid_Kp, DEC);
    Serial.print('\t');
    Serial.print(pid_Ki, DEC);
    Serial.print('\t');
    Serial.print(pid_Kd, DEC);
    Serial.print('\n');
  }
}

void ReadSerial()
{
  while (Serial.available() > 0)
  {
    switch(Serial.read())
    {
      case 'p':
        pid_Kp = Serial.parseFloat();
        break;
      case 'i':
        pid_Ki = Serial.parseFloat();
        break;
      case 'd':
        pid_Kd = Serial.parseFloat();
        break;
      case 't':
        temp_target = Serial.parseFloat();
        target_update = true;
        break;
    }
  }  
}

void PIcontroller()
{
  if (ds_update)
  {
    if (temp_target < dewpoint) {
      temp_target_limited = dewpoint; 
      peltier_mode = 0;
    }
    else {
      temp_target_limited = temp_target;
      peltier_mode = 2;
    }
    
    float error = temp_target_limited - temp_cmos;

    if (abs(error) > 0.5)
      peltier_mode = 1;

    // calculate integral part
    pid_integral += error;
    if (pid_integral > pid_integral_max) pid_integral = pid_integral_max; // prevent windup
    if (pid_integral < pid_integral_min) pid_integral = pid_integral_min;

    float pid_derivative = pid_error_last - error;

    float out = error*pid_Kp + pid_integral*pid_Ki; + pid_derivative*pid_Kd;
    if (out > max_duty) out = max_duty;
    if (out < 0.) out = 0.;
    
    peltier_duty = out*100.;
    
    byte pwm_out = 255 - (byte)(out*255);
    
    analogWrite(PELTIER_PIN,pwm_out);
  }
}

void Encoder_int()
{  
  if(digitalRead(ENCODER_SA_PIN))
  {
    temp_target += Encoder_step;
  }
  else
  {  
    temp_target -= Encoder_step;
  }     
  target_update = true;
}
